@@ -1,9 +1,12 @@
+import codecs
 import pathlib
 from pprint import pformat
 from typing import Optional, Sequence
 
 import typer
 
+from openapi_python_client import MetaType
+from openapi_python_client.config import Config
 from openapi_python_client.parser.errors import ErrorLevel, GeneratorError, ParseError
 
 app = typer.Typer()
@@ -17,14 +20,13 @@ def _version_callback(value: bool) -> None:
         raise typer.Exit()
 
 
-def _process_config(path: Optional[pathlib.Path]) -> None:
-    from .config import Config
+def _process_config(path: Optional[pathlib.Path]) -> Config:
 
     if not path:
-        return
+        return Config()
 
     try:
-        Config.load_from_path(path=path)
+        return Config.load_from_path(path=path)
     except:  # noqa
         raise typer.BadParameter("Unable to parse config")
 
@@ -33,11 +35,8 @@ def _process_config(path: Optional[pathlib.Path]) -> None:
 @app.callback(name="openapi-python-client")
 def cli(
     version: bool = typer.Option(False, "--version", callback=_version_callback, help="Print the version and exit"),
-    config: Optional[pathlib.Path] = typer.Option(
-        None, callback=_process_config, help="Path to the config file to use"
-    ),
 ) -> None:
-    """ Generate a Python client from an OpenAPI JSON document """
+    """Generate a Python client from an OpenAPI JSON document"""
     pass
 
 
@@ -56,7 +55,7 @@ def _print_parser_error(e: GeneratorError, color: str) -> None:
 
 
 def handle_errors(errors: Sequence[GeneratorError]) -> None:
-    """ Turn custom errors into formatted error messages """
+    """Turn custom errors into formatted error messages"""
     if len(errors) == 0:
         return
     error_level = ErrorLevel.WARNING
@@ -104,14 +103,24 @@ custom_template_path_options = {
     "resolve_path": True,
 }
 
+_meta_option = typer.Option(
+    MetaType.POETRY,
+    help="The type of metadata you want to generate.",
+)
+
+CONFIG_OPTION = typer.Option(None, "--config", help="Path to the config file to use")
+
 
 @app.command()
 def generate(
     url: Optional[str] = typer.Option(None, help="A URL to read the JSON from"),
     path: Optional[pathlib.Path] = typer.Option(None, help="A path to the JSON file"),
     custom_template_path: Optional[pathlib.Path] = typer.Option(None, **custom_template_path_options),  # type: ignore
+    meta: MetaType = _meta_option,
+    file_encoding: str = typer.Option("utf-8", help="Encoding used when writing generated"),
+    config_path: Optional[pathlib.Path] = CONFIG_OPTION,
 ) -> None:
-    """ Generate a new OpenAPI Client library """
+    """Generate a new OpenAPI Client library"""
     from . import create_new_client
 
     if not url and not path:
@@ -120,7 +129,22 @@ def generate(
     if url and path:
         typer.secho("Provide either --url or --path, not both", fg=typer.colors.RED)
         raise typer.Exit(code=1)
-    errors = create_new_client(url=url, path=path, custom_template_path=custom_template_path)
+
+    try:
+        codecs.getencoder(file_encoding)
+    except LookupError:
+        typer.secho("Unknown encoding : {}".format(file_encoding), fg=typer.colors.RED)
+        raise typer.Exit(code=1)
+
+    config = _process_config(config_path)
+    errors = create_new_client(
+        url=url,
+        path=path,
+        meta=meta,
+        custom_template_path=custom_template_path,
+        file_encoding=file_encoding,
+        config=config,
+    )
     handle_errors(errors)
 
 
@@ -129,8 +153,11 @@ def update(
     url: Optional[str] = typer.Option(None, help="A URL to read the JSON from"),
     path: Optional[pathlib.Path] = typer.Option(None, help="A path to the JSON file"),
     custom_template_path: Optional[pathlib.Path] = typer.Option(None, **custom_template_path_options),  # type: ignore
+    meta: MetaType = _meta_option,
+    file_encoding: str = typer.Option("utf-8", help="Encoding used when writing generated"),
+    config_path: Optional[pathlib.Path] = CONFIG_OPTION,
 ) -> None:
-    """ Update an existing OpenAPI Client library """
+    """Update an existing OpenAPI Client library"""
     from . import update_existing_client
 
     if not url and not path:
@@ -140,5 +167,19 @@ def update(
         typer.secho("Provide either --url or --path, not both", fg=typer.colors.RED)
         raise typer.Exit(code=1)
 
-    errors = update_existing_client(url=url, path=path, custom_template_path=custom_template_path)
+    try:
+        codecs.getencoder(file_encoding)
+    except LookupError:
+        typer.secho("Unknown encoding : {}".format(file_encoding), fg=typer.colors.RED)
+        raise typer.Exit(code=1)
+
+    config = _process_config(config_path)
+    errors = update_existing_client(
+        url=url,
+        path=path,
+        meta=meta,
+        custom_template_path=custom_template_path,
+        file_encoding=file_encoding,
+        config=config,
+    )
     handle_errors(errors)
